#include <cmath>
#include <cstdlib>
#include "camera.h"
#include "light.h"
#include "matrix.h"
#include "misc.h"
#include "renderer.h"
#include "scene.h"
#include "shadowmap.h"
#include "tests.h"
#include "texunit.h"

using namespace std;

namespace Msp {
namespace GL {

ShadowMap::ShadowMap(unsigned s, const Renderable &r, const Light &l):
	Effect(r),
	size(s),
	light(l),
	radius(1),
	depth_bias(4),
	rendered(false)
{
	depth_buf.set_min_filter(LINEAR);
	depth_buf.set_compare_enabled(true);
	depth_buf.set_compare_func(LEQUAL);
	depth_buf.set_wrap(CLAMP_TO_EDGE);
	depth_buf.storage(DEPTH_COMPONENT, size, size);
	fbo.attach(DEPTH_ATTACHMENT, depth_buf, 0);
	fbo.require_complete();

	set_darkness(0.7);
}

void ShadowMap::set_target(const Vector3 &t, float r)
{
	target = t;
	radius = r;
}

void ShadowMap::set_darkness(float d)
{
	if(d<0.0f || d>1.0f)
		throw invalid_argument("ShadowMap::set_darkness");

	shdata.uniform("shadow_darkness", d);
}

void ShadowMap::set_depth_bias(float b)
{
	if(b<0.0f)
		throw invalid_argument("ShadowMap::set_depth_bias");

	depth_bias = b;
}

void ShadowMap::setup_frame() const
{
	if(rendered)
		return;

	rendered = true;
	renderable.setup_frame();

	Camera camera;
	const Vector4 &lpos = light.get_position();
	/* XXX Not really proper way to support positional lights, but good
	enough when the light source is far away */
	camera.set_look_direction(lpos.w*target-lpos.slice<3>(0));

	camera.set_up_direction((abs(camera.get_look_direction().z)<0.99) ? Vector3(0, 0, 1) : Vector3(0, 1, 0));
	camera.set_position(target);
	camera.set_orthographic(radius*2, radius*2);
	camera.set_depth_clip(-radius, radius);

	shadow_matrix = camera.get_object_matrix();
	shadow_matrix.scale(radius*2, radius*2, -radius*2);
	shadow_matrix.translate(-0.5, -0.5, depth_bias/size-0.5);
	shadow_matrix.invert();

	BindRestore bind_fbo(fbo);
	Bind bind_depth(DepthTest::lequal());
	fbo.clear(DEPTH_BUFFER_BIT);

	Renderer renderer(&camera);
	renderable.render(renderer, "shadow");
}

void ShadowMap::finish_frame() const
{
	renderable.finish_frame();
	rendered = false;
}

void ShadowMap::render(Renderer &renderer, const Tag &tag) const
{
	if(!enabled_passes.count(tag))
		return renderer.render(renderable, tag);

	Renderer::Push _push_rend(renderer);

	unsigned unit = renderer.allocate_effect_texunit();
	int iunit = unit;
	shdata.uniform("shadow", iunit);

	Bind _bind_depth(depth_buf, unit);

	if(const Camera *camera = renderer.get_camera())
		/* Multiply by camera's object matrix to form a matrix that transforms
		from eye space to shadow space. */
		shdata.uniform("shd_eye_matrix", shadow_matrix*camera->get_object_matrix());
	else
		shdata.uniform("shd_eye_matrix", shadow_matrix);

	renderer.add_shader_data(shdata);
	renderer.render(renderable, tag);
}

} // namespace GL
} // namespace Msp
