#include <algorithm>
#include <msp/strings/format.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include "error.h"
#include "vertexformat.h"

using namespace std;

namespace Msp {
namespace GL {

VertexFormat::VertexFormat():
	data(0)
{ }

VertexFormat::VertexFormat(VertexComponent c):
	data(new unsigned char[8])
{
	data[0] = 1;
	data[1] = c;
}

VertexFormat::VertexFormat(const VertexFormat &f):
	data(0)
{
	*this = f;
}

VertexFormat &VertexFormat::operator=(const VertexFormat &f)
{
	delete[] data;
	if(f.data)
	{
		data = new unsigned char[(f.data[0]&~7)+8];
		copy(f.data, f.data+f.data[0]+1, data);
	}
	else
		data = 0;

	return *this;
}

VertexFormat::~VertexFormat()
{
	delete[] data;
}

VertexFormat VertexFormat::operator,(VertexComponent c) const
{
	VertexFormat r = *this;
	if(r.data)
	{
		const unsigned char n = ++r.data[0];
		if((n&7)==0)
		{
			unsigned char *newdt = new unsigned char[n+8];
			copy(r.data, r.data+n, newdt);
			delete r.data;
			r.data = newdt;
		}
		r.data[n] = c;
	}
	else
	{
		r.data = new unsigned char[8];
		r.data[0] = 1;
		r.data[1] = c;
	}

	return r;
}

VertexFormat VertexFormat::operator,(unsigned i) const
{
	if(!data)
		throw invalid_operation("VertexFormat::operator,");
	VertexFormat r = *this;
	unsigned char *c = r.data+r.data[0];
	*c = make_indexed_component(static_cast<VertexComponent>(*c), i);

	return r;
}

unsigned VertexFormat::stride() const
{
	unsigned s = 0;
	for(const unsigned char *i=begin(); i!=end(); ++i)
		s += get_component_size(*i);
	return s;
}

int VertexFormat::offset(VertexComponent comp) const
{
	unsigned type = get_component_type(comp);
	unsigned size = get_component_size(comp);
	unsigned offs = 0;
	for(const unsigned char *i=begin(); i!=end(); ++i)
	{
		if(get_component_type(*i)==type)
		{
			if(get_component_size(*i)>=size)
				return offs;
			else
				return -1;
		}
		else
			offs += get_component_size(*i);
	}

	return -1;
}

VertexComponent make_indexed_component(VertexComponent comp, unsigned index)
{
	if(comp>=TEXCOORD1 && comp<=TEXCOORD4)
	{
		if(index>=4)
			throw out_of_range("make_indexed_component");
	}
	else if(comp>=ATTRIB1 && comp<=ATTRIB4)
	{
		if(index>=24)
			throw out_of_range("make_indexed_component");
	}
	else
		throw invalid_argument("make_indexed_component");
	return static_cast<VertexComponent>(comp+index*4);
}

void operator>>(const LexicalConverter &conv, VertexComponent &c)
{
	const string &str = conv.get();
	if(str.size()==7 && !str.compare(0, 6, "VERTEX") && str[6]>='2' && str[6]<='4')
		c = static_cast<VertexComponent>(VERTEX2+(str[6]-'2'));
	else if(str=="NORMAL3")
		c = NORMAL3;
	else if(str.size()==12 && !str.compare(0, 5, "COLOR") && str[5]>='3' && str[5]<='4' && !str.compare(6, 6, "_FLOAT"))
		c = static_cast<VertexComponent>(COLOR3_FLOAT+(str[5]-'3'));
	else if(str=="COLOR4_UBYTE")
		c = COLOR4_UBYTE;
	else if(str=="TANGENT3")
		c = TANGENT3;
	else if(str=="BINORMAL3")
		c = BINORMAL3;
	else if(str.size()>=9 && !str.compare(0, 8, "TEXCOORD") && str[8]>='1' && str[8]<='4')
	{
		if(str.size()==9)
			c = static_cast<VertexComponent>(TEXCOORD1+(str[8]-'1'));
		else if(str.size()==11 && str[9]=='_' && str[10]>='0' && str[10]<='7')
			c = static_cast<VertexComponent>(TEXCOORD1+(str[8]-'1')+(str[10]-'0')*4);
		else
			throw lexical_error(format("conversion of '%s' to VertexComponent", str));
	}
	else if(str.size()>=9 && !str.compare(0, 6, "ATTRIB") && str[6]>='1' && str[6]<='4' && str[7]=='_')
	{
		unsigned n;
		try
		{
			n = lexical_cast<unsigned>(str.substr(8));
		}
		catch(const lexical_error &)
		{
			throw lexical_error(format("conversion of '%s' to VertexComponent", str));
		}
		c = static_cast<VertexComponent>(ATTRIB1+(str[6]-'1')+n*4);
	}
	else
		throw lexical_error(format("conversion of '%s' to VertexComponent", str));
}

} // namespace GL
} // namespace Msp
